#pragma once

#include "Mesh.h"
#include "Shader.h"
#include "Renderable.h"

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/matrix_decompose.hpp>

/**
 * @brief A renderable body with a mesh visualization.
 */
class Body : public Renderable
{
public:
    /**
     * @brief Construct a new Body object based on an .obj-file.
     * @param meshPath String that represents the file location to load.
     */
    Body( const char *meshPath );

    /**
     * @brief Destroy the Body object
     */
    virtual ~Body();

    /**
     * @brief Generates the GPU memory to render the separating planes.
     */
    void generateSepPlanesBuffers();

    /**
     * @brief Render the body's mesh with the given shader.
     * @param shader The shader to render the body's mesh with.
     */
    virtual void render( Shader &shader ) override;

    /**
     * @brief Set the position, as well as update the transformation matrix.
     * @param translation The new position of the body in world space.
     */
    void setTranslation( glm::vec3 translation ) { m_position = translation; m_modelToWorld_translation = glm::translate(translation); updateModelToWorld(); };

    /**
     * @brief Set the rotation, as well as update the transformation matrix.
     * @param angle The angle of the new rotation of the body in degrees.
     * @param axis The axis of the new rotation of the body in world space.
     */
    void setRotation( float angle, glm::vec3 axis ) { m_modelToWorld_rotation = glm::rotate(angle,axis); updateModelToWorld(); };

    /**
     * @brief Set the rotation, as well as update the transformation matrix.
     * @param rotation The new rotation of the body in world space.
     */
    void setRotation( glm::mat4 rotation ) { m_modelToWorld_rotation = rotation; updateModelToWorld(); };

    /**
     * @brief Set the scale, as well as update the transformation matrix.
     * @param scale The new scale of the body in world space.
     */
    void setScale( glm::vec3 scale ) { m_modelToWorld_scale = glm::scale(scale); updateModelToWorld(); };

    /** 
     * @brief Directly set the model to world matrix.
     * @note Also tries to update the transformation parts by decomposing
     *       the combined transformation in parts. Reliability unclear.
    */
    void setModelToWorld( glm::mat4 WfM ) {
        glm::vec3 scale;
        glm::quat rotation;
        glm::vec3 translation;
        glm::vec3 skew;
        glm::vec4 perspective;
        glm::decompose(WfM, scale, rotation, translation, skew, perspective);
        m_modelToWorld = WfM;
        m_modelToWorld_rotation = glm::mat4(glm::conjugate(rotation));
        m_modelToWorld_translation = glm::translate(translation);
        m_position = translation;
        m_modelToWorld_scale = glm::scale(scale);
    };

    /**
     * @brief Get the body's mesh vertex at index.
     * @param index The index of the vertex inside the mesh's vertex list.
     * @return Vertex at index.
     */
    Vertex getVertex( unsigned int index ) { return m_mesh.vertices[index]; }
    
    /**
     * @brief Get the vertex count of the body's mesh.
     * @return The vertex count of the body's mesh.
     */
    size_t getVertexCount() { return m_mesh.vertices.size(); }

    /**
     * @brief Get the position of the body's origin in world space.
     * @return The position of the body's origin in world space.
     */
    glm::vec3 getPosition() { return m_position; };
    
    /**
     * @brief Get the translational part of the transformation.
     * @return The translational part of the transformation.
     */
    glm::mat4 getTranslation() { return m_modelToWorld_translation; };

    /**
     * @brief Get the rotational part of the transformation.
     * @return The rotational part of the transformation.
     */
    glm::mat4 getRotation() { return m_modelToWorld_rotation; };

    /**
     * @brief Get the scaling part of the transformation.
     * @return The scaling part of the transformation.
     */
    glm::mat4 getScale() { return m_modelToWorld_scale; };

    /**
     * @brief Get the Model To World transformation.
     * @return The model to world transformation.
     * @ref m_modelToWorld for what this transformation does.
     */
    glm::mat4 getModelToWorld() { return m_modelToWorld; };

    /** The body's color in RGBA in the range [0,1]^4. */
    glm::vec4 m_color = glm::vec4( 1.0 );

    /** Whether this body's sep. plane is valid. */
    bool m_sepPlaneValid = true;

    /**
     * @brief This body's separating plane in world space.
     * 
     * The fourth component represents the signed distance of the
     * plane to the origin (also in world space). The first three
     * components represent the plane's normal, which is
     * perpendicular to the surface of the plane.
     * 
     * @note Pay attention in what space the separating plane
     *       is stored in, the rendering code in PlaneCube::render
     *       expects the planes in world space. Make sure to
     *       store them in world space, or transform them
     *       accordingly in PlaneCube::render.
     */
    glm::vec4 m_sepPlane = glm::vec4(0.333f, 0.5f, 0.1, 1);

    /** Vertex Array Object for rendering the separating plane. */
    GLuint m_sepPlanes_vao = 0;

    /** Vertex Buffer Object for rendering the separating plane. */
    GLuint m_sepPlanes_vbo = 0;

    /** Element Array Object for rendering the separating plane. */
    GLuint m_sepPlanes_ebo = 0;

private:
    /** The mesh that is rendered at the body's position. */
    Mesh m_mesh;

    /** The translational part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_translation = glm::mat4( 1.0 );

    /** The rotational part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_rotation = glm::mat4( 1.0 );

    /** The scaling part of the body's world from model transformation. */
    glm::mat4 m_modelToWorld_scale = glm::mat4( 1.0 );

    /**
     * @brief The body's world from model transformation.
     * 
     * When multiplied with a point in model space (such as given in an
     * .obj-file) from the left side, the point will be transformed in
     * to world space, based on the current position of the body.
     */
    glm::mat4 m_modelToWorld = glm::mat4( 1.0 );

    /** The position of the body's origin in world space. */
    glm::vec3 m_position = glm::vec3( 0 );

    /**
     * @brief Calculates the world from model transformation based the
     *        current member values of the transformation parts.
     * @note This might override a previously directly written modelToWorld
     *       matrix, as the parts of the transformation might not have been
     *       extracted properly.
     */
    void updateModelToWorld() { m_modelToWorld = m_modelToWorld_translation * m_modelToWorld_rotation * m_modelToWorld_scale; };
};
