#include "Mesh.h"

#include <iostream>

// We use the tiny_obj_loader library to load data of .obj-files into this mesh:
#define TINYOBJLOADER_IMPLEMENTATION

// We want to use triangulation when loading non-triangulated polygons:
#define TINYOBJLOADER_USE_MAPBOX_EARCUT

// Include tiny_obj_loader:
#include "tiny_obj_loader.h"

Mesh::Mesh()
{
}

Mesh::Mesh(std::vector<Triangle> triangles) : triangles(triangles) {
    createBuffers();
}

void Mesh::createBuffers() {
    glGenBuffers(1, &vbo);
    glGenBuffers(1, &vio);

    std::vector<Vertex> vboData;
    unsigned int n = 0;
    for (; n < triangles.size();) {
        for (int i = 0; i < 2; ++i) {
            vboData.push_back(*(&triangles[n].a + i + (i % 2 == 0 ? 1 : -1)));
        }
        vboData.push_back(*(&triangles[n].c));
        ++n;
    }

    std::vector<unsigned int> indices;
    for (unsigned int i = 0; i < vboData.size(); ++i) {
        indices.push_back(i % 3 == 0 ? i + 1 : i % 3 == 1 ? i - 1 : i);
    }

    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vio);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), &indices[0], GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, vboData.size() * sizeof(Vertex), &vboData[0], GL_STATIC_DRAW);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), 0);
    glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(sizeof(glm::vec4)));
    glVertexAttribPointer(2, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(2 * sizeof(glm::vec4)));
    glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), reinterpret_cast<void*>(3 * sizeof(glm::vec4)));

    for (int i = 0; i < 4; ++i)
        glEnableVertexAttribArray(i);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}

/* This contructor loads an obj file into the VAO and VBO (see header) */
Mesh::Mesh(std::string filepath){
    tinyobj::attrib_t attrib;
    std::vector<tinyobj::shape_t> shapes;
    std::vector<tinyobj::material_t> materials;
    std::string warn, err;
    bool retval = tinyobj::LoadObj( &attrib, &shapes, &materials, &warn, &err, filepath.c_str() );
    if (!warn.empty())
        std::cout << "[Mesh] Warning in LoadObj: " << warn << std::endl;
    if (!err.empty())
        std::cout << "[Mesh] Error in LoadObj: " << err << std::endl;
    if (!retval)
        std::cout << "[Mesh] LoadObj failed!" << std::endl;
    // std::cout << "[Mesh] Loading " << filepath << std::endl;

    std::vector<glm::vec4> vertices_;
    std::vector<glm::vec4> normals;
    for (int i=0; i+2<attrib.vertices.size(); i+=3)
        vertices_.push_back( glm::vec4(attrib.vertices[i+0],
                                       attrib.vertices[i+1],
                                       attrib.vertices[i+2],
                                       1 ) );
    for (int i=0; i+2<attrib.normals.size(); i+=3)
        normals.push_back( glm::normalize(
                             glm::vec4(attrib.normals[i+0],
                                       attrib.normals[i+1],
                                       attrib.normals[i+2],
                                       0 )) );
    for (int i=0; i<shapes.size(); ++i) {
        for (int j=0; j+2<shapes[i].mesh.indices.size(); j+=3) {
            tinyobj::index_t i0 = shapes[i].mesh.indices[j+0];
            tinyobj::index_t i1 = shapes[i].mesh.indices[j+1];
            tinyobj::index_t i2 = shapes[i].mesh.indices[j+2];
            Triangle triangle;
            triangle.a.position = vertices_[i0.vertex_index];
            triangle.b.position = vertices_[i1.vertex_index];
            triangle.c.position = vertices_[i2.vertex_index];
            triangle.a.normal = normals[i0.normal_index];
            triangle.b.normal = normals[i1.normal_index];
            triangle.c.normal = normals[i2.normal_index];
            triangles.push_back( triangle );
        }
    }

    // Map vertex neighbours, iterate over all vertices
    for (int i=0; i<triangles.size(); ++i) {
        auto center_triangle = &triangles[i];
        for (int j=0; j<3; ++j) {
            auto center_vertex = &(*center_triangle)[j];

            // Per vertex, find all triangles that it is connected to
            std::vector<Triangle*> neighbour_triangles;
            for (int k=0; k<triangles.size(); ++k) {
                auto other_triangle = &triangles[k];
                if (other_triangle->isConnectedTo( center_vertex ))
                    neighbour_triangles.push_back( other_triangle );
            }

            // Create neighbouring vertex list based on found connected triangles
            for (int k=0; k<neighbour_triangles.size(); ++k) {
                auto neighbour_triangle = neighbour_triangles[k];
                for (int l=0; l<3; ++l) {
                    auto neighbour_vertex = &(*neighbour_triangle)[l];
                    if (glm::length(neighbour_vertex->position-center_vertex->position) > FLT_EPSILON) {
                        bool contained = false;
                        for (int m=0; m<center_vertex->neighbours.size(); ++m) {
                            auto containedVertex = center_vertex->neighbours[m];
                            if (glm::length(containedVertex->position-center_vertex->position) < FLT_EPSILON)
                                contained = true;
                        }
                        if (!contained)
                            center_vertex->neighbours.push_back( neighbour_vertex );
                    }
                }
            }
        }
    }

    // Create unique list of vertices
    for (int i=0; i<triangles.size(); ++i) {
        auto triangle = &triangles[i];
        for (auto vertex : {&triangle->a, &triangle->b, &triangle->c}) {
            bool contained = false;
            for (auto containedVertex : vertices) {
                if (glm::length(containedVertex.position - vertex->position) < FLT_EPSILON) {
                    contained = true;
                    break;
                }
            }
            if (!contained) {
                vertices.push_back( *vertex );
            }
        }
    }

    createBuffers();
}

Mesh::~Mesh(){
    // If there is a vao, delete it from the GPU:
    if(vao != 0)
        glDeleteVertexArrays(1, &vao);

    // Same for the vbo:
    if(vbo != 0)
        glDeleteBuffers(1, &vbo);

    // Same for the vio:
    if(vio != 0)
        glDeleteBuffers(1, &vio);

    // To avoid reusage after deletion:
    vao = 0;
    vbo = 0;
    vio = 0;
}

void Mesh::render( Shader &shader )
{
    shader.bind();

    // Draws the geometry data that is defined by the vao:
    glBindVertexArray(vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vio);
    glDrawElements(GL_TRIANGLES, (GLsizei)(triangles.size()*3), GL_UNSIGNED_INT, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
}
