// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include this file only once when compiling:
#pragma once

// Include the std::vector class, which is comparable to the
// ArrayList in Java (this is not a mathematical vector!):
#include <vector>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include matrix and vector:
#include <glm/glm.hpp>

#include <string>

// Include our Triangle primitive:
#include "Triangle.h"

#include "Renderable.h"
#include "Shader.h"

/**
 * Represents a renderable mesh.
 */
class Mesh : public Renderable
{
public:
    Mesh();

    /**
     * Constructs a mesh which consists of the given vertices
     * and prepare for rendering (e.g. upload the data to the GPU).
     */
    Mesh(std::vector<Triangle> triangles);

    /**
     * Loads the geometry data of an .obj file into this mesh object.
     */
    Mesh(std::string filepath);

    /**
     * Deletes the mesh and ensures that the memory on the GPU
     * is cleaned up.
     */
    ~Mesh();

    /**
     * Renders the mesh.
     */
    virtual void render( Shader &shader ) override;

    /**
     * Creates buffers (VAO,VBO,etc.) at the GPU and uploads triangle
     * data. In previous frameworks, this was done directly in the
     * constructor, but it was splitted to be use with different
     * constructors.
     */
    void createBuffers();

    // Pointer to Vertex Array Object (on GPU):
    GLuint vao = 0;

    // Pointer to vertex buffer (interleaved):
    GLuint vbo = 0;
    GLuint vio = 0;

    std::vector<Triangle> triangles;
    std::vector<Vertex> vertices;
};
