#include "PlaneCube.h"
#include "Body.h"

#define GLM_FORCE_RADIANS
#include <glm/glm.hpp>
#include <glm/gtx/transform.hpp>

#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

#include <glad/glad.h>
#include <GLFW/glfw3.h>

#define CLIPPING 1

static const float AabbExtend = 4.f;

PlaneCube::PlaneCube()
{
    // The vertices of the unit cube.
    static const std::vector<glm::vec4> vertices_ = {
        glm::vec4(-1,-1, 1,1), // 0
        glm::vec4( 1,-1, 1,1), // 1
        glm::vec4(-1, 1, 1,1), // 2
        glm::vec4( 1, 1, 1,1), // 3
        glm::vec4(-1,-1,-1,1), // 4
        glm::vec4( 1,-1,-1,1), // 5
        glm::vec4(-1, 1,-1,1), // 6
        glm::vec4( 1, 1,-1,1), // 7
    };

    // Construct auxiliary plane objects to perform 3D
    // clipping of the separating planes for visualization.
    Plane px; px.normal = glm::vec3( 1, 0, 0); px.distance = -AabbExtend*(1); m_clippingPlanes.push_back(px);
    Plane py; py.normal = glm::vec3( 0, 1, 0); py.distance = -AabbExtend*(1); m_clippingPlanes.push_back(py);
    Plane pz; pz.normal = glm::vec3( 0, 0, 1); pz.distance = -AabbExtend*(1); m_clippingPlanes.push_back(pz);
    Plane nx; nx.normal = glm::vec3(-1, 0, 0); nx.distance = -AabbExtend*(1); m_clippingPlanes.push_back(nx);
    Plane ny; ny.normal = glm::vec3( 0,-1, 0); ny.distance = -AabbExtend*(1); m_clippingPlanes.push_back(ny);
    Plane nz; nz.normal = glm::vec3( 0, 0,-1); nz.distance = -AabbExtend*(1); m_clippingPlanes.push_back(nz);

    // The indices to render the lines of the unit cube as
    // degenerate triangles (only 2 points are unique, 1
    // is a duplicate).
    static const std::vector<GLuint> indices {
        // Top quad
        2, 6, 6,
        2, 3, 3,
        3, 7, 7,
        7, 7, 6,

        // Bottom quad
        0, 4, 4,
        0, 1, 1,
        1, 5, 5,
        5, 5, 4,

        // Left lines
        0, 0, 2,
        3, 3, 1,

        // Right lines
        4, 4, 6,
        5, 5, 7,
    };

    // Generate triangle objects for the unit cube
    for (int i=0; i<indices.size(); i+=3)
        triangles.push_back( Triangle(vertices_[indices[i+0]],
                                      vertices_[indices[i+1]],
                                      vertices_[indices[i+2]]) );

    // Call Mesh::createBuffers to create the GPU buffers based
    // on the previously setup vertices and indices.
    createBuffers();
}

PlaneCube::PlaneCube( std::shared_ptr<Scene> pScene )
    : PlaneCube()
{
    scene = pScene;
}

void PlaneCube::updateSeparatingPlanes() const
{
    if (!scene)
        return;

    // *** Separating Planes Algorithm ***

    auto P = scene->centerBody;

    for (int i=0; i<scene->orbitingBodies.size(); ++i) {
        auto Q = scene->orbitingBodies[i];
        auto H = Q->m_sepPlane;

        // TODO: Separating Planes
        // Perform the separating planes algorithm with spatio-
        // temporal coherence. Spatio-temporal coherence
        // involves using the most recent sep. plane as the
        // starting candidate, as well as using steepest
        // descent to quickly find the next "offending" vertex.
        //
        // Note: Pay attention in what space the vertices
        // are, as they are stored in object space and the
        // separating planes should be stored in world space.
        // Keeping them in world space will make it easier
        // to reuse as well as render them later on.
        // You can transform a vertex from model / object space
        // to world space by multiplying it from the right side
        // to the body's model-to-world matrix.
        // In any case, plane and vertex must to be in the same
        // space to (correctly) perform certain operations on
        // them such as +.
        //
        // Note: Pay attention to signs (+/-) as one of the
        // object, for example P, should be on the positive
        // side of the plane (H*p>0) and the other object
        // should be on the negative side (H*q<0).
        // Therefore, you also need to adjust the learning
        // formular H=H+eta*p for q.
    }
}

void addToQuad(std::vector<glm::vec3> &oldQuad, glm::vec3 newVertex)
{
    if (!oldQuad.empty())
        oldQuad.pop_back();
    oldQuad.push_back( newVertex );
    oldQuad.push_back( oldQuad[0] );
}

std::vector<glm::vec3> PlaneCube::SutherlandHodgman( std::vector<glm::vec3> quad, std::vector<Plane> clippingPlanes )
{
    static const float Epsilon = 0.0001f;
    float d1, d2;
    std::vector<glm::vec3> result_quad = quad;
    for (int i=0; i<clippingPlanes.size(); ++i) {
        if (result_quad.empty())
            return result_quad;
        auto plane = clippingPlanes[i];
        std::vector<glm::vec3> clipped_quad;
        auto points = result_quad;
        for (int j=0; j<points.size()-1; ++j) {
            auto p = points[j];
            auto pn = points[j+1];
            d1 = plane.distanceTo(p);
            d2 = plane.distanceTo(pn);

            if (d1<=0 && d2<=0)
                addToQuad( clipped_quad, pn );
            else if ((d1>0) && ((d2 > -Epsilon) && (d2 < Epsilon)))
                addToQuad( clipped_quad, pn );
            else if ((d1>-Epsilon && d1<Epsilon) && (d2>0))
                continue;
            else if (d1<=0 && d2>0)
                addToQuad( clipped_quad, plane.intersectionPoint(p,pn) );
            else if (d1>0 && d2<=0) {
                auto newp = plane.intersectionPoint(p, pn);
                addToQuad( clipped_quad, newp );
                addToQuad( clipped_quad, pn );
            }
        }
        result_quad = clipped_quad;
    }
    return result_quad;
}

void PlaneCube::render( Shader &shader )
{
    // Separating planes
    auto t0 = std::chrono::high_resolution_clock::now();
    updateSeparatingPlanes();
    auto t1 = std::chrono::high_resolution_clock::now();

    // Exponential moving average of timing
    auto dto = m_sepPlaneDurationMicroSecEma;
    auto dtn = ((float)(t1-t0).count())/1000.f;
    static const float EmaRatio = 0.1f;
    m_sepPlaneDurationMicroSecEma = (1-EmaRatio)*dto + EmaRatio*dtn;
    m_sepPlaneDurationMicroSec = dtn;
    if (m_frameCounter%60 == 0) {
        m_fps = (unsigned int)std::round((1000.f*1000.f)/m_sepPlaneDurationMicroSecEma);
    }

    // Prepare shader uniforms
    shader.bind();
    shader.setUniform( "objectColor", glm::vec4(1,1,1,1) );
    shader.setUniform( "model", glm::scale(glm::vec3(AabbExtend)) );

    // Draw wireframe cube
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    glBindVertexArray(vao);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vio);
    // GL_LINES might be easier to use here
    glDrawElements(GL_TRIANGLES, (GLsizei)(triangles.size()*3), GL_UNSIGNED_INT, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindVertexArray(0);
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

    if (!scene)
        return;

    // We render in world space from now on
    shader.setUniform( "model", glm::identity<glm::mat4>() );


    // Iterate over bodies and render separating plane
    for (int i=0; i<scene->orbitingBodies.size(); ++i) {
        auto Q = scene->orbitingBodies[i];

        // Check if the separating plane is intersection-free
        if (Q->m_sepPlaneValid) {
            // Set a semi-transparent color based on body color
            static const float Alpha = 0.5f;
            auto color = glm::vec4(glm::vec3(Q->m_color), Alpha);
            shader.setUniform( "objectColor", color );

            // Construct a plane space
            auto plane = Q->m_sepPlane;
            auto normal = glm::normalize(glm::vec3(plane));
            auto distance = plane.w;
            auto point = normal * -distance/glm::length(glm::vec3(plane));
            glm::vec3 pz = normal;
            glm::vec3 px;
            if (std::abs(normal.x) < std::abs(normal.y))
                px = glm::cross(normal, glm::vec3(1.f, 0.f, 0.f));
            else
                px = glm::cross(normal, glm::vec3(0.f, 1.f, 0.f));
            auto py = glm::normalize(glm::cross(pz, px));

            // Triangulate plane
#if CLIPPING
            static const float PlaneExtend = 100.f;
#else
            static const float PlaneExtend = 2.f;
#endif
            glm::vec3 ext(PlaneExtend);
            glm::vec3 vertices[4];
            vertices[0] = point +  px*PlaneExtend +  py*PlaneExtend; // Left Top
            vertices[1] = point +  px*PlaneExtend + -py*PlaneExtend; // Left Bottom
            vertices[2] = point + -px*PlaneExtend + -py*PlaneExtend; // Right Bottom
            vertices[3] = point + -px*PlaneExtend +  py*PlaneExtend; // Right Top

            // Clip the plane on the cube bounds
            std::vector<glm::vec3> quad;
            quad.insert(quad.end(), &vertices[0], &vertices[4]);
            quad.push_back(vertices[0]);
#if CLIPPING
            auto q = SutherlandHodgman( quad, m_clippingPlanes );
#else
            auto q = quad;
#endif

            // Render a semi-transparent plane
            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            glBindVertexArray(Q->m_sepPlanes_vao);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, Q->m_sepPlanes_ebo);
            glBindBuffer(GL_ARRAY_BUFFER, Q->m_sepPlanes_vbo);
            glBufferData(GL_ARRAY_BUFFER, q.size()*sizeof(glm::vec3), &q[0], GL_STATIC_DRAW);
            // GL_TRIANGLE_FAN handily constructs a convex polygon from all vertices
            glDrawElements(GL_TRIANGLE_FAN, (GLsizei)(q.size()-1), GL_UNSIGNED_INT, 0);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
            glBindVertexArray(0);
            glDisable(GL_BLEND);
        }
    }

    ++m_frameCounter;
}
