#pragma once

#include "Renderable.h"
#include "Mesh.h"
#include "Triangle.h"
#include "Scene.h"

#include <thread>
#include <chrono>

class PlaneCube : public Mesh
{
public:
    /**
     * @brief Represents a plane at p = normal * distance.
     */
    struct Plane {
        glm::vec3 normal;
        float distance;
        float distanceTo(glm::vec3 p) {
            return glm::dot(normal,p) + distance;
        }
        glm::vec3 intersectionPoint(glm::vec3 p1, glm::vec3 p2) {
            return p1 + (p2-p1) * (-distanceTo(p1)/glm::dot(normal,p2-p1));
        }
    };

    /**
     * @brief Construct a new Plane Cube object.
     */
    PlaneCube();

    /**
     * @brief Construct a new Plane Cube object with scene pointer.
     * @param scene The scene object, the pointer is stored for later access.
     */
    PlaneCube( std::shared_ptr<Scene> scene );

    /**
     * @brief Clip a polygon based on Sutherland Hodgman Algorithm.
     * @param quad The input polygon.
     * @param clippingPlanes The clipping planes.
     * @return std::vector<glm::vec3> The polygon 3D clipped on all clipping planes.
     */
    static std::vector<glm::vec3> SutherlandHodgman( std::vector<glm::vec3> quad, std::vector<Plane> clippingPlanes );

    /**
     * @brief Update the separating planes of the scene
     * 
     * Go through all oribiting bodies and update its separating plane
     * with spatio-temporal coherence. Spatial coherence is achieved
     * by storing the last frames separating plane and using that as
     * the initial candidate and using steepest descent to find the
     * next vertex on the wrong side of the sep. plane in the least
     * amount of steps (this is only possible because of convexity).
     */
    void updateSeparatingPlanes() const;

    /**
     * @brief Render a cube and all valid separating planes.
     * @param shader The shader to use for rendering.
     */
    virtual void render( Shader &shader ) override;

    /** The scene pointer to access other bodies. */
    std::shared_ptr<Scene> scene;

    /** Auxiliary variable to clip visual planes */
    std::vector<Plane> m_clippingPlanes;

    /** The last sep. planes frametime in microseconds. */
    float m_sepPlaneDurationMicroSec = 0.f;

    /** The last frametime after moving average smoothing. */
    float m_sepPlaneDurationMicroSecEma = 0.f;

    /** The FPS based on the current smoothed frametime (60 Hz update). */
    unsigned int m_fps = 0;

    /** The frame counter. */
    unsigned int m_frameCounter = 1;
};