// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include this file only once when compiling:
#pragma once

// Include utility to open files during runtime:
#include <fstream>
#include <iostream>

// Include string:
#include <string>

// Include matrix and vector:
#include <glm/glm.hpp>

/**
 * This is a wrapper class for shader programs in OpenGL.
 */

class Shader {
public:
    /** Stores the ID of the shader program on the GPU */
    unsigned int shaderProgram;

    /** Is set to true in the constructor if initialization worked: */
    bool initialized = false;

public:
    /**
     * Loads the source code of the given file paths, compiles it on the
     * GPU and stores the reference (=id, =name) to the program in the
     * attribute 'shaderProgram'.
     *
     * NOTE:
     * OpenGL shaders are usually not compiled when the C++ program
     * is compiled, but only when the program is started. I.e. we have
     * to read in from the source files at this point and then pass it
     * to OpenGL to be compiled on the graphics card.
     *
     * In recent OpenGL versions, it is also possible to compile shaders
     * before hand (so they don't have to be compiled every time you
     * start a game, but that's another topic ;-)).
     */
    Shader(std::string vertexShaderPath, std::string fragmentShaderPath);

    /**
     * Deletes the shader and ensures that the memory on the GPU
     * is cleaned up.
     */
    ~Shader();

    /**
     * Binds the shader so that it will be used in future draw calls.
     */
    void bind();

    /**
     * Sets a uniform variable of type Mat4f.
     */
    void setUniform(std::string name, glm::mat4 value);

    /**
     * Sets a uniform variable of type Vec4f.
     */
    void setUniform(std::string name, glm::vec4 value);

    /**
     * Sets a uniform variable of type Vec3f.
     */
    void setUniform(std::string name, glm::vec3 value);

    /**
     * Sets a uniform variable of type float.
     */
    void setUniform(std::string name, float value);

    /**
     * Sets a uniform variable of type int.
     */
    void setUniform(std::string name, int value);
};
